#!/usr/bin/perl
##*****************************************************************************
##
##     File Name:	DBFind.pl
##         Owner:	InfoPlace
##     Copyright:	InfoPlace Inc., Pleasanton, CA, USA, 1996.
##     				Not to be copied, modified or redistributed without written
##     				permission of InfoPlace.
##
##
##*****************************************************************************
##                      R E V I S I O N   H I S T O R Y
##*****************************************************************************
##
##	  Date		Time	Author	Description
##	--------	-----	------	---------------------------------------------
##	09/15/96	12:51	N		Original version
##
##*****************************************************************************


require "DBOpen.pl";
require "DBIterate.pl";
require "SList.pl";

# --------------------------------------------------------------------------------
# CompareStrings($str1, $str2, $op)
# Compares string 1 with string 2 using the operator specified.
# Args:
# $str1	(string): string 1
# $str2	(string): string 2
# $op	(integer): 0: equal, -1: not equal
#				   1:partial string with wild card prefix (*str)
#				   2:partial string with wild card suffix (str*)
#				   3:partial string with wild card padded (*str*)
# Return Value	(boolean): 0 if compare failed, 1 if compare sucessful.
#
sub CompareStrings {
	local($str1, $str2, $op) = @_;
	my($escstr2) = $str2;
	$escstr2 =~ s/([^a-zA-Z0-9])/\\$1/g;
	#&LogMsg("Compare String: v1='" . $str1 . "' v2='" . $str2 ."' op=" . $op);
	if ( $op == 0 ) {
		return ((($str1 eq $str2) || ($str1 =~ /^(\s*)$escstr2(\s*)$/i)) ? 1 : 0);
	}
	if ( $op == -1 ) {
		return (( $str1 ne $str2 ) ? 1 : 0);
	}
	if ( $op == 1 ) {
		return (( $str1 =~ /(.*)$escstr2$/i ) ? 1 : 0);
	}
	if ( $op == 2 ) {
		return (( $str1 =~ /^$escstr2(.*)/i ) ? 1 : 0);			
	}
	if ( $op == 3 ) {
		return (( $str1 =~ /(.*)$escstr2(.*)/i ) ? 1 : 0);			
	}
	return 0;
}

# --------------------------------------------------------------------------------
# CompareNumbers($num1, $num2, $op)
# Compares number 1 with number 2 using the operator specified.
# Args:
# $num1	(string): number 1
# $num2	(string): number 2
# $op	(integer): 0: equal, -1: not equal
#				   1:greater than, 2: less than
# Return Value	(boolean): 0 if compare failed, 1 if compare sucessful.
#
sub CompareNumbers {
	local($num1, $num2, $op) = @_;
	if ( $op == 0 ) {
		return (( $num1 == $num2 ) ? 1 : 0);
	}
	if ( $op == -1 ) {
		return (( $num1 != $num2 ) ? 1 : 0);
	}
	if ( $op == 1 ) {
		return (( $num1 > $num2 ) ? 1 : 0);
	}
	if ( $op == 2 ) {
		return (( $num1 < $num2 ) ? 1 : 0);			
	}
	return 0;
}

# --------------------------------------------------------------------------------
# CompareLists($values, $criteria, $type, $compareType)
# Checks if the list of values (defined as a string, seperated by "," char)
# matches the criterion specified, type defines the value type and 
# compareType defines how to compare.
#

sub CompareLists
{
	local($values, $criteria, $type, $compareType) = @_;
	local(@vlist) = split($kValueSeperator, $values);
	local(@clist) = split($kValueSeperator, $criteria);
	local($vcount) = $#vlist + 1;
	local($ccount) = $#clist + 1;
	local($i, $j);
	local($result) = 0;
	#&LogMsg("Compare List: v1='" . $values . "' v2='" . $criteria ."' type=" . $type . " compare=" . $compareType);
	for ($j = 0; $j < $ccount && !$result; $j++) {
		for ($i = 0; $i < $vcount && !$result; $i++) {
			if ($type == $kStringType) {
				$result = &CompareStrings($vlist[$i], $clist[$j], $compareType);
			}
			else {
				if ($type == $kNumericType)  {
					$result = &CompareNumbers($vlist[$i], $clist[$j], $compareType);
				}
			}
		}
	}
	#&LogMsg("Result: " . $result);
	return $result;
}

# --------------------------------------------------------------------------------
# CompareValues($value, $criterion, $type, $compareType)
# Checks if the value matches the criterion specified, type defines the 
# value type and compareType defines how to compare.
#
sub CompareValues {
	local($value, $criterion, $type, $compareType) = @_;
	local($typeIndex) = $kDBFieldTypeNameIndexAssoc{$type};
	if ($kStringType ==  $typeIndex || $kUnknowntype == $typeIndex) {
		return &CompareStrings($value, $criterion, $compareType);
	}
	if ($kNumericType ==  $typeIndex) {
		return &CompareNumbers($value, $criterion, $compareType);
	}
	if ($kListOfStringType ==  $typeIndex) {
		return &CompareLists($value, $criterion, $kStringType, $compareType);
	}
	if ($kListOfNumericType ==  $typeIndex) {
		return &CompareLists($value, $criterion, $kNumericType, $compareType);
	}
	return 0;
}

# --------------------------------------------------------------------------------
# DBMatchRecord(@record, @fieldTypes, @criteria, @compareTypes)
# Checks for a match for the given record according to the given criteria.
# Args:
#	@record	(list): category record value as a list. Here the fields not 
#						needing to be matched should have 'undefined' as value.
#						Following fields may be specified:
#						(name, description, parentCategory) 
#	@fieldTypes (list): a list of field types for the given record.
#	@criteria (list): a list of document field values specified as criteria.
#	@compareTypes (list): each element indicates how to compare corresponding field
#						 value specified in criteria list.
#			for all values:- 0: equal, -1: not equal
#			for string values:- 1:partial string with wild card prefix (*str)
#							    2:partial string with wild card suffix (str*)
#			for numeral values:- 1:greater than, 2: less than
#
#	Return Value(boolean): 0 if find successful, 1 if not.

sub DBMatchRecord
{
	local(*record, *fieldTypes, *criteria, *compareTypes, $orConditions) = @_;
	local($i, $criteriaCount, $result) = (0, $#criteria + 1, 1);
	if ($criteriaCount == 0) {
		return 0;
	}
	#&LogMsg("SearchRecord: " . join("|", @record));
	if ($orConditions) {
		$result = 0;
		while (($i < $criteriaCount) && !$result) {
			if (defined($criteria[$i])) {
				$result = &CompareValues($record[$i], $criteria[$i], $fieldTypes[$i], $compareTypes[$i]);
			}
			$i++;
		}
	}
	else {
		while (($i < $criteriaCount) && $result) {
			if (defined($criteria[$i])) {
				$result = &CompareValues($record[$i], $criteria[$i], $fieldTypes[$i], $compareTypes[$i]);
			}
			$i++;
		}
	}
	#&LogMsg("SearchRecord Result: $result");
	return $result;
}

# --------------------------------------------------------------------------------
# DBFindRecords(@db, @criteriaValue, @compareTypes, @records)
# Returns the a list of records that match the specified criteria.
# Args:
#	@db			(list): the db to search from
#	@criteriaValue	(list): record value as a list. Here the fields not needing 
#						to be matched should have 'undefined' as value.
#	@compareTypes (list): each element indicates how to compare corresponding field
#						 value specified in criteria list.
#			for all values:- 0: equal, -1: not equal
#			for string values:- 1:partial string with wild card prefix (*str)
#							    2:partial string with wild card suffix (str*)
#							    3:partial string with wild card padded (*str*)
#			for numeral values:- 1:greater than, 2: less than
#	@records	(list): list of matched records (records are in string form)
#	Return Value(boolean): 0 if find successful, 1 if not.
sub DBFindRecords
{
	local(*db, *criteriaValue, *compareTypes, *records, $orConditions) = @_;
	local($result);
	local(@record) = ();
	local(@fieldTypes) = ();
	$result = &DBFieldTypes(*db, *fieldTypes);
	if ($result == $errNone) {
		local($i) = 0;
		$result = &DBReadFirstRecord(*db, *record);
		while ($result == $errNone) {
			if (&DBMatchRecord(*record, *fieldTypes, *criteriaValue, *compareTypes, $orConditions)) {
				### $records[$i++] = &SListMake(*record, $kFieldSeperator);
				$records[$i++] = $record[0]; # just the key
			}
			$result = &DBReadNextRecord(*db, *record);
		}
		$result = $errNone;
	}
	return $result;
}
