# $Id: DocAPI.pl,v 1.1 1998/05/04 19:50:02 naga Exp $
# Copyright InfoPlace Inc., Milpitas, CA, USA, 1996-1998
# Not to be copied, modified or redistributed without written
# permission of InfoPlace.
#
# --------------------------------------------------------------------------------

require "LogMsg.pl";
require "CGIUtility.pl";
require "DBOpen.pl";
require "DBLookUp.pl";
require "DBRemove.pl";
require "DBIterate.pl";
require "DBAdd.pl";
require "SList.pl";
require "DocUtility.pl";

# --------------------------------------------------------------------------------
# OPAddCategory($name, $parentID, $desc)
# Adds a category record with specified $name under optionally specified parent category.
# Args:
#	$name 		(string): the name of the category.
#	$parentID   (numeric): the id of the parent category.
#	$desc		(string): a description for the category.
#							
#	Return Value(boolean): 0 if add successful, $errDuplicateCategory if already 
#							category by the same name exists, $errParentDoesNotExist
#							if parent does not exists, with "" as special case,
#                           $errCategoryInParent, if category is already in the parent,
#							$errOPFailed otherwise
sub OPAddCategory
{
	my($name, $parentID, $desc) = @_;
	my($result);
	my(@db) = ();
	
	if (!defined($parentID) || $parentID == 0 || $parentID eq "") {
		$parentID = $kRootCategoryID;
	}
	$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBWriteAccess, \@db);
	
	# acquire write lock
	open(LOCKFILE, ">$kCatDocLockFile");
	&GetWriteLockOnFile("LOCKFILE") || &LogMsg("GetWriteLockOnFile failed on file handle $kCatDocLockFile: $!\n");
	
	if ($result == $errNone) {
		my(@parentRecord);
		$result = &DBRetreiveRecord(\@db, $parentID, \@parentRecord);
		if ($result == 2) {
			&DBClose(\@db);
			return $errParentDoesNotExist;
		}
		if ($result == $errNone) {
			my($children) = $parentRecord[$kCategoryChildListIndex];
			my($catid);
			if (&IsCategoryInList(\@db, $name, $children, \$catid)) {
				return $errCategoryInParent;  # already category by same name exists in the parent category
			}
			my($currentTime) = time;
			my($id) = &GetAndIncrementMasterID($kMasterCatIDFile);
			my(@record) = ($id, $name, $parentID, "", "", $desc, $currentTime, $currentTime, 0);
			$result = &DBAppendRecord(\@db, \@record);
			if ($result == $errNone) {
				$children = &SListAppend($children, $id, $kValueSeperator);
				$parentRecord[$kCategoryChildListIndex] = $children;
				$parentRecord[$kCategoryModificationTimeIndex] = $currentTime;
				&DBReplaceRecord(\@db, $parentID, \@parentRecord);
				&OPAddToCategoryMap($name, $id);
			}
		}
		&DBClose(\@db);
	}
	# release write lock
	close(LOCKFILE);
	&UnlockFile("LOCKFILE") || &LogMsg("UnlockFile failed on file handle $kCatDocLockFile: $!\n");
	
	return $result;
}

# --------------------------------------------------------------------------------
# OPAddToCategoryMap($name, $id)
# Args:
#	$name 		(string): the name of the category.
#	$id        (numeric): the id of the category.
#							
#	Return Value(boolean): 0 if add successful, $errOPFailed otherwise
sub OPAddToCategoryMap
{
	my($name, $id) = @_;
	my($result);
	my(@db) = ();
	$result = &DBOpen($kCategoryMapDB, $kDocDBPath, $kDBWriteAccess, \@db);
	
	if ($result == $errNone) {
		my(@record) = ();
		$result = &DBRetreiveRecord(\@db, $name, \@record);
		if ($result != 2 && $result != 0) {
			&DBClose(\@db);
			return $errOPFailed;
		}
		if ($result == 2) {
			$record[$kCategoryMapNameIndex] = $name;
			$record[$kCategoryMapIDListIndex] = $id;
			$result = &DBAppendRecord(\@db, \@record);
		}
		else {
			$record[$kCategoryMapIDListIndex] = &SListAppend($record[$kCategoryMapIDListIndex], $id, $kValueSeperator);
			$result = &DBReplaceRecord(\@db, $name, \@record);
		}
	}
	&DBClose(\@db);
	return $result;
}

# --------------------------------------------------------------------------------
# OPRemoveFromCategoryMap($name, $id)
# Args:
#	$name 	    (string): the name of the category.
#	$id        (numeric): the id of the category.
#							
#	Return Value(boolean): 0 if add successful, $errOPFailed otherwise
sub OPRemoveFromCategoryMap
{
	my($name, $id) = @_;
	my($result);
	my(@db) = ();
	$result = &DBOpen($kCategoryMapDB, $kDocDBPath, $kDBWriteAccess, \@db);
	
	if ($result == $errNone) {
		my(@record) = ();
		$result = &DBRetreiveRecord(\@db, $name, \@record);
		if ($result != 0) {
			&DBClose(\@db);
			return $errOPFailed;
		}
		$record[$kCategoryMapIDListIndex] = &SListRemove($record[1], $id, $kValueSeperator);
		if (!$record[$kCategoryMapIDListIndex] || !$record[$kCategoryMapIDListIndex] eq "") {
			$result = &DBRemoveRecord(\@db, $name);
		}
		else {
			$result = &DBReplaceRecord(\@db, $name, \@record);
		}
	}
	&DBClose(\@db);
	return $result;
}

# --------------------------------------------------------------------------------
# OPAddDocument($name, $categories, @docRecord)
# Adds a document record with specified $name under optionally specified parent category.
# Args:
#	$name 		(string): the name of the document.
#	$categories (string): the name of the parent categories, names are seperated 
#                         by $kValueSeperator
#	@docRecord	(list): (author, url, size, appType, versionID, previousVersionDocID, 
#						keyWordlist, desc)
#							
#	Return Value(boolean): 0 if add successful, 1 if not.
sub OPAddDocument
{
    local($name, $categoriesAsString, *docRecord) = @_;
    my($result, $i, $parentCount, $category);
    my(@db) = ();
    $result = &DBOpen($kDocumentDB, $kDocDBPath, $kDBWriteAccess, \@db);

	# acquire write lock
	open(LOCKFILE, ">$kCatDocLockFile");
	&GetWriteLockOnFile("LOCKFILE") || &LogMsg("GetWriteLockOnFile failed on file handle $kCatDocLockFile: $!\n");

    if ($result == $errNone) {
		my(@categories) = &SListUnmake($categoriesAsString, $kValueSeperator);
		$parentCount = $#categories + 1;
		my(@catdb) = ();
		$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBWriteAccess, \@catdb);
		my(@catrecord) = ();
		for ($i = 0; $i < $parentCount; $i++) {
			$category = $categories[$i];
			$result = &DBRetreiveRecord(\@catdb, $category, \@catrecord);
			if ($result == 2) {
				&DBClose(\@catdb);
				&DBClose(\@db);
				&LogMsg("Document Creation: Parent: $category does not exist");
				return $errParentDoesNotExist;
			}
		}
		my($currentTime) = time;
		my($docID) = &GetAndIncrementMasterID($kMasterDocIDFile);
		my(@record) = ($docID, $name, $docRecord[0], $docRecord[1], $currentTime,
					   $currentTime, $docRecord[4], $docRecord[2], 
					   $docRecord[3], $categoriesAsString, $docRecord[5], 0, 0, $docRecord[6], $docRecord[7]);
		
		$result = &DBAppendRecord(\@db, \@record);
		&DBClose(\@db);
		if( $result != 0 ) {
			return $result;
		}
		# add url to urlid map ...
		$result = &OPAddDocIdToURLIdMap($docID, $record[$kDocURLIndex], 0);
		
		# hook up the doc to the category ...
		for ($i = 0; $i < $parentCount; $i++) {
			$category = $categories[$i];
			$result = &DBRetreiveRecord(\@catdb, $category, \@catrecord);
			my($docs) = $catrecord[$kCategoryDocIDListIndex];
			$docs = &SListAppend($docs, $docID, $kValueSeperator);
			$catrecord[$kCategoryDocIDListIndex] = $docs;
			$catrecord[$kCategoryModificationTimeIndex] = $currentTime;
			&DBReplaceRecord(\@catdb, $category, \@catrecord);
		}
		&DBClose(\@catdb);	   
		return $result;
    }
    # release write lock
	close(LOCKFILE);
	&UnlockFile("LOCKFILE") || &LogMsg("UnlockFile failed on file handle $kCatDocLockFile: $!\n");

    return $errOPFailed;
}

# --------------------------------------------------------------------------------
# OPGetCategory($key, @category)
# Retreive the category record with specified $key (only specified field values are filled in).
# Args:
#	$key 		(numeric): the id of the category.
#	@category	(list): category record returned as a list.

#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPGetCategory
{
	local($key, *category) = @_;
	my($result);
	my(@db) = ();
	$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBReadAccess, \@db);
	if ($result == $errNone) {
		$result = &DBRetreiveRecord(\@db, $key, *category);
		&DBClose(\@db);
	}
	return $result;
}

# --------------------------------------------------------------------------------
# OPGetDocument($key, @document)
# Retreive the document record with specified $key
# Args:
#	$key 		(string): the key for the document record.
#	@document	(list): document record returned as a list.					

#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPGetDocument
{
	local($key, *document) = @_;
	my($result);
	my(@db) = ();
	$result = &DBOpen($kDocumentDB, $kDocDBPath, $kDBReadAccess, \@db);
	if ($result == $errNone) {
		$result = &DBRetreiveRecord(\@db, $key, *document);
		&DBClose(\@db);
	}
	return $result;
}

# --------------------------------------------------------------------------------
# OPSetCategory($key, @category)
# Sets the category record for specified $key.
# Args:
#	$key 	(numeric): the id of the category.
#	@category	(list): category record value as a list. Here the fields not 
#						needing to be set should have 'undefined' as value.

#	Return Value(boolean): 0 if set successful, 1 if not.
sub OPSetCategory
{
	local($key, *category) = @_;
	my($result);
	my(@recordz) = ();
	my(@db) = ();
	$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBWriteAccess, \@db);

    # acquire write lock
	open(LOCKFILE, ">$kCatDocLockFile");
	&GetWriteLockOnFile("LOCKFILE") || &LogMsg("GetWriteLockOnFile failed on file handle $kCatDocLockFile: $!\n");
	
	$result = &DBRetreiveRecord(\@db, $key, \@recordz);

	# root is a special case ... its name and parents properties CANNOT be changed
	if ($key eq $kRootCategoryID) {
		$category[$kCategoryNameIndex] = $recordz[$kCategoryNameIndex];
		$category[$kCategoryParentIndex] = "";
	}

	if ($result == $errNone) {
		my($currentTime) = time;
		$category[$kCategoryModificationTimeIndex] = $currentTime;
		my($i, $count);
		$count = $#category + 1;
		my($ignoreNameChange) = 1;
		my($ignoreParentChange) = 1;
		my($oldName) = $recordz[$kCategoryNameIndex];
		my($oldParent) = $recordz[$kCategoryParentIndex];
		# we start from 1, since 0, is the $key ...
		for ($i = 1; $i < $count; $i++) {
			if (defined($category[$i])) {
				if ($i == $kCategoryNameIndex) {
					$ignoreNameChange = ($oldName eq $category[$i]);
				}
				else {
					if ($i == $kCategoryParentIndex) {
						$ignoreParentChange = ($oldParent == $category[$i]);
					}
				}
				$recordz[$i] = $category[$i];
			}
		}
		if (!$ignoreNameChange) {
			$result = &OPRemoveFromCategoryMap($oldName, $key);
			$result = &OPAddToCategoryMap($category[$kCategoryNameIndex], $key);
			if ($result != $errNone) {
				# error state ..
				return 2;
			}
		}
		$result = &DBReplaceRecord(\@db, $key, \@recordz);
		if (!$ignoreParentChange) {
			my(@parentRecord) = ();
			$result = &DBRetreiveRecord(\@db, $oldParent, \@parentRecord);
			my($children) = $parentRecord[$kCategoryChildListIndex];
			$children = &SListRemove($children, $key, $kValueSeperator);
			$parentRecord[$kCategoryChildListIndex] = $children;
			$result = &DBReplaceRecord(\@db, $oldParent, \@parentRecord);
			my($newParent) = $recordz[$kCategoryParentIndex];
			$result = &DBRetreiveRecord(\@db, $newParent, \@parentRecord);
			$children = $parentRecord[$kCategoryChildListIndex];
			$children = &SListAppend($children, $key, $kValueSeperator);
			$parentRecord[$kCategoryChildListIndex] = $children;
			$result = &DBReplaceRecord(\@db, $newParent, \@parentRecord);
		}
		&DBClose(\@db);
	}
	# release write lock
	close(LOCKFILE);
	&UnlockFile("LOCKFILE") || &LogMsg("UnlockFile failed on file handle $kCatDocLockFile: $!\n");

	return $result;
}

# --------------------------------------------------------------------------------
# OPSetDocument($key, @document)
# Sets the category record for specified $key.
# Args:
#	$key 		(string): the key for the document record.
#	@document	(list): document record value as a list. Here the fields not 
#						needing to be set should have 'undefined' as value.

#	Return Value(boolean): 0 if set successful, 1 if not.
sub OPSetDocument
{
	local($key, *document) = @_;
	my($i, $count);
	my($currentTime) = time;
	$document[$kDocModificationTimeIndex] = $currentTime;
	my($result);

	# acquire write lock
	open(LOCKFILE, ">$kCatDocLockFile");
	&GetWriteLockOnFile("LOCKFILE") || &LogMsg("GetWriteLockOnFile failed on file handle $kCatDocLockFile: $!\n");

	my(@record) = ();
	my(@db) = ();
	$result = &DBOpen($kDocumentDB, $kDocDBPath, $kDBWriteAccess, \@db);
	$result = &DBRetreiveRecord(\@db, $key, \@record);
	if ($result == $errNone) {
		if ($document[$kDocURLIndex] && $document[$kDocURLIndex] ne $record[$kDocURLIndex]) { 
			# if url is changed ...
			$result = &OPAddDocIdToURLIdMap($key, $document[$kDocURLIndex], 0);
			$result = &OPRemoveDocIdFromURLIdMap($key, $record[$kDocURLIndex], 0);
			
		}
		my($ignoreParentChange) = 1;
		my($oldParents) = $record[$kDocCategoryListIndex];
		$count = $#document + 1;
		for ($i = 0; $i < $count; $i++) {
			if (defined($document[$i])) { 
				if ($i == $kDocCategoryListIndex) {
					$ignoreParentChange = ($oldParents eq $document[$i]);
				}
				$record[$i] = $document[$i];
			}
		}
		$result = &DBReplaceRecord(\@db, $key, \@record);
		&DBClose(\@db);
		if (!$ignoreParentChange) {
			$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBWriteAccess, \@db);
			my(@parentRecord) = ();
			my($parent, $i, $count);
			$count = &SListLength($oldParents, $kValueSeperator);
			for ($i = 0; $i < $count; $i++) {
				$parent = &SListNthElement($oldParents, $i, $kValueSeperator);
				$result = &DBRetreiveRecord(\@db, $parent, \@parentRecord);
				my($children) = $parentRecord[$kCategoryDocIDListIndex];
				$children = &SListRemove($children, $key, $kValueSeperator);
				$parentRecord[$kCategoryDocIDListIndex] = $children;
				$parentRecord[$kCategoryModificationTimeIndex] = $currentTime;
				$result = &DBReplaceRecord(\@db, $parent, \@parentRecord);
			}
			my($newParents) = $record[$kDocCategoryListIndex];
			$count = &SListLength($newParents, $kValueSeperator);
			for ($i = 0; $i < $count; $i++) {
				$parent = &SListNthElement($newParents, $i, $kValueSeperator);
				$result = &DBRetreiveRecord(\@db, $parent, \@parentRecord);
				my($children) = $parentRecord[$kCategoryDocIDListIndex];
				$children = &SListAppend($children, $key, $kValueSeperator);
				$parentRecord[$kCategoryDocIDListIndex] = $children;
				$parentRecord[$kCategoryModificationTimeIndex] = $currentTime;
				$result = &DBReplaceRecord(\@db, $parent, \@parentRecord);
			}
			&DBClose(\@db);
		}
	}
	else {
		return $errEntryNotFound;
	}
    # release write lock
	close(LOCKFILE);
	&UnlockFile("LOCKFILE") || &LogMsg("UnlockFile failed on file handle $kCatDocLockFile: $!\n");

	return $result;
}

# --------------------------------------------------------------------------------
# OPGetDocumentVersions($key, @documentList)
# Retreive the document records for each version of the document with specified $key
# Args:
#	$key 			(string): the key for the document record.
#	@documentList	(list): document records returned as a list.					

#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPGetDocumentVersions
{
	local($key, *documentList) = @_;
	my($result);
	my(@record) = ();

	$result = &OPGetDocument($key, \@record);
	if ($result != $errNone) {
		return $result;
	}
	$documentList[$record[$kDocVersionIndex] - 1] = &SListMake(\@record, $kFieldSeperator);
	my($nextDocID) = $record[$kDocNextVersionIDIndex];
	my($previousDocID) = $record[$kDocPreviousVersionIDIndex];
	while ($nextDocID) {
		$result = &OPGetDocument($nextDocID, \@record);
		if ($result != $errNone) {
			return $result;
		}
		$documentList[$record[$kDocVersionIndex] - 1] = &SListMake(\@record, $kFieldSeperator);
		$nextDocID = $record[$kDocNextVersionIDIndex];
	}
	while ($previousDocID) {
		$result = &OPGetDocument($nextDocID, \@record);
		if ($result != $errNone) { 
			return $result;
		}
		$documentList[$record[$kDocVersionIndex] - 1] = &SListMake(\@record, $kFieldSeperator);
		$previousDocID = $record[$kDocPreviousVersionIDIndex];
	}
	return 0;
}

# --------------------------------------------------------------------------------
# OPGetAllCategories(@categoryList)
# Retreive all the category records (only specified field values are filled in).
# Args:
#	@categoryList 	(list): category records returned as a list.

#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPGetAllCategories
{
	local(*categoryList) = @_;
	my($result);
	my(@db) = ();
	$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBReadAccess, \@db);
	if ($result == $errNone) {
		my(@category) = ();
		$result = &DBReadFirstRecord(\@db, \@category);
		my($i) = 0;
		while (!$result) {
			$categoryList[$i++] = &SListMake(\@category, $kFieldSeperator);
			$result = &DBReadNextRecord(\@db, \@category);
		}
	}
	&DBClose(\@db);
	return $result;	
}

sub SortNamesWithID
{
	local(*names, *ids) = @_;
	my($done) = 0;
	my($size) = $#names;
	my($i);
	while (!$done) {
		$done = 1;
		for ($i = 0; $i < $size; $i++) {
			my($t1) = $names[$i];
			my($t2) =  $names[$i+1];
			my($tt1) = $t1;
			my($tt2) = $t2;
			$tt1 =~ tr/A-Z/a-z/;
			$tt2 =~ tr/A-Z/a-z/;
			my($result) =  $tt1 cmp  $tt2;
			if ($result > 0) {
				$done = 0;
				$names[$i] = $t2;
				$names[$i+1] = $t1;
				$t1 = $ids[$i];
				$t2 = $ids[$i + 1];
				$ids[$i] = $t2;
				$ids[$i+1] = $t1;
			}
		}
	}
}

# --------------------------------------------------------------------------------
sub OPGetCategoryNamesFromIDs
{
	local(*parents, *names) = @_;
	my($count) = $#parents + 1;
	my($i);
	for ($i = 0; $i < $count; $i++) {
		my(@record) = ();
		&OPGetCategory($parents[$i], \@record);
		$names[$i] = $record[$kCategoryNameIndex];
	}
}


# --------------------------------------------------------------------------------
# OPGetAllCategoriesName(@idList, @nameList, $sortit, $type)
# Retreive all the categories' name 
# Args:
#	@idList 	(list): categories id, returned as a list.
#	@nameList 	(list): categories name, returned as a list.
#	$dosort 	(boolean): indicates, if the names need to be sorted
#	$type 	        (1, 2, 3): 1 implies all categories, 
#                                  2 implies categories with docs only
#                                  3 implies categories with categories only

#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPGetAllCategoriesName
{
	local(*idList, *nameList, $doSort, $type) = @_;
	my($result);
	my(@db) = ();
	$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBReadAccess, \@db);
	if ($result == $errNone) {
		my(@category) = ();
		$result = &DBReadFirstRecord(\@db, \@category);
		my($i) = 0;
		while (!$result) {
		    my($docs) = $category[$kCategoryDocIDListIndex];
		    my($cats) = $category[$kCategoryChildListIndex];
		    if (!$type || $type == 1 || (($type == 2) && !$cats) || (($type == 3) && !$docs)) {
				$idList[$i] = $category[$kCategoryIDIndex];
				$nameList[$i++] = $category[$kCategoryNameIndex];
		    }
		    $result = &DBReadNextRecord(\@db, \@category);
		}
		if ($doSort) {
			&SortNamesWithID(*nameList, *idList);
		}
	}
	&DBClose(\@db);
	return $result;	
}

# --------------------------------------------------------------------------------
# OPRemoveCategory($key)
# Remove the category record with specified $key.
# Args:
#	$key 		(string): the key for the category record.
#   $lock       (undefined or 0 or 1): this args should always be undefined or 0
#               [this is used internally as this function is recursive]
#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPRemoveCategory
{
	my($key, $lock) = @_;
	my($result);
	my(@db) = ();
	my(@record);
	$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBWriteAccess, \@db);

	if (!$lock) {
		# acquire write lock
		open(LOCKFILE, ">$kCatDocLockFile");
		&GetWriteLockOnFile("LOCKFILE") || &LogMsg("GetWriteLockOnFile failed on file handle $kCatDocLockFile: $!\n");
	}

	$result = $result || &DBRetreiveRecord(\@db, $key, \@record);
	if ($result == $errNone) {
		# detach self from the parent ...
		my($parent) = $record[$kCategoryParentIndex];
		my(@parentRecord) = ();
		$result = &DBRetreiveRecord(\@db, $parent, \@parentRecord);
		my($children) = $parentRecord[$kCategoryChildListIndex];
		$children = &SListRemove($children, $key, $kValueSeperator);
		$parentRecord[$kCategoryChildListIndex] = $children;
		my($currentTime) = time;
		$parentRecord[$kCategoryModificationTimeIndex] = $currentTime;
		$result = &DBReplaceRecord(\@db, $parent, \@parentRecord);
		
		# remove children documents ...
		$children = $record[$kCategoryDocIDListIndex];
		my($count, $i, $doc);
		$count = &SListLength($children, $kValueSeperator);
		for ($i = 0; $i < $count; $i++) {
			$doc = &SListNthElement($children, $i, $kValueSeperator);
			$result = &OPRemoveDocumentFromCategory($doc, $key, 0);
		}
		# remove children categories ...
		$children = $record[$kCategoryChildListIndex];
		my($category);
		$count = &SListLength($children, $kValueSeperator);
		for ($i = 0; $i < $count; $i++) {
			$category = &SListNthElement($children, $i, $kValueSeperator);
			$result = &OPRemoveCategory($category, 1);
		}
		# finally remove self ...
		$result = &DBRemoveRecord(\@db, $key);
		&DBClose(\@db);
	}
	if (!$lock) {
		# release write lock
		close(LOCKFILE);
		&UnlockFile("LOCKFILE") || &LogMsg("UnlockFile failed on file handle $kCatDocLockFile: $!\n");
	}
	return $result;
}

# --------------------------------------------------------------------------------
# OPRemoveDocument($key)
# Remove the document record with specified $key.
# Args:
#	$key 		(string): the key for the document record.

#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPRemoveDocument
{
    my($key) = @_;
    my($result);
    my(@db) = ();
    my(@record);
    $result = &DBOpen($kDocumentDB, $kDocDBPath, $kDBWriteAccess, \@db);

	# acquire write lock
	open(LOCKFILE, ">$kCatDocLockFile");
	&GetWriteLockOnFile("LOCKFILE") || &LogMsg("GetWriteLockOnFile failed on file handle $kCatDocLockFile: $!\n");

    $result = $result || &DBRetreiveRecord(\@db, $key, \@record);
    #$result = $result || &OPRemoveDocIdFromURLIdMap($key, $record[$kDocURLIndex], 0);
	#donot check the return value, as there is a problem with url character set, some funny char
	#screws up the remove docid-url map, this in turns fails remove document itself
    &OPRemoveDocIdFromURLIdMap($key, $record[$kDocURLIndex], 0);
    if ($result == $errNone) {
		my($parents) = $record[$kDocCategoryListIndex];
		my($parent, $i, $count);
		my(@parentRecord, @dbCategory) = ((), ());
		$count = &SListLength($parents, $kValueSeperator);
		$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBWriteAccess, \@dbCategory);
		for ($i = 0; $i < $count; $i++) {
			$parent = &SListNthElement($parents, $i, $kValueSeperator);
			$result = &DBRetreiveRecord(\@dbCategory, $parent, \@parentRecord);
			my($children) = $parentRecord[$kCategoryDocIDListIndex];
			$children = &SListRemove($children, $key, $kValueSeperator);
			$parentRecord[$kCategoryDocIDListIndex] = $children;
			my($currentTime) = time;
			$parentRecord[$kCategoryModificationTimeIndex] = $currentTime;
			$result = &DBReplaceRecord(\@dbCategory, $parent, \@parentRecord);
		}
		$result = &DBRemoveRecord(\@db, $key);
		&DBClose(\@db);
    }

    # release write lock
	close(LOCKFILE);
	&UnlockFile("LOCKFILE") || &LogMsg("UnlockFile failed on file handle $kCatDocLockFile: $!\n");

    return $result;
}

# --------------------------------------------------------------------------------
# OPReplaceParentInDocument($docid, $oldparent, $newparent)
sub OPReplaceParentInDocument
{
    my($docid, $oldparent, $newparent) = @_;
    my($result);
    my(@db) = ();
    my(@record);
    $result = &DBOpen($kDocumentDB, $kDocDBPath, $kDBWriteAccess, \@db);
    $result = $result || &DBRetreiveRecord(\@db, $docid, \@record);
    if ($result == $errNone) {
	my($parents) = $record[$kDocCategoryListIndex];
	$parents = &SListReplace($parents, $oldparent, $newparent, $kValueSeperator);
	$record[$kDocCategoryListIndex] = $parents;
	$result = &DBReplaceRecord(\@db, $docid, \@record);
    }
    &DBClose(\@db);
    return $result;    
}

# --------------------------------------------------------------------------------
# OPRemoveDocumentFromCategory($key, $category)
# Remove the document record with specified $key.
# Args:
#	$key 		(string): the key for the document record.
#	$category 	(string): the parent category from which the doc needs to be removed

#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPRemoveDocumentFromCategory
{
	my($key, $category, $updateCategory) = @_;
	my($result);
	my(@db) = ();
	my(@record);
	$result = &DBOpen($kDocumentDB, $kDocDBPath, $kDBWriteAccess, \@db);
	if ($result == $errNone) {
		if ($updateCategory) {
			my(@dbCategory) = ();
			my(@parentRecord) = ();
			$result = &DBOpen($kCategoryDB, $kDocDBPath, $kDBWriteAccess, \@dbCategory);
			$result = $result || &DBRetreiveRecord(\@dbCategory, $category, \@parentRecord);
			if ($result == $errNone) {
				my($children) = $parentRecord[$kCategoryDocIDListIndex];
				$children = &SListRemove($children, $key, $kValueSeperator);
				$parentRecord[$kCategoryDocIDListIndex] = $children;
				my($currentTime) = time;
				$parentRecord[$kCategoryModificationTimeIndex] = $currentTime;
				$result = &DBReplaceRecord(\@dbCategory, $category, \@parentRecord);
			}
			&DBClose(\@dbCategory);
		}
		$result = $result || &DBRetreiveRecord(\@db, $key, \@record);
		if ($result == $errNone) {
			my($parents) = $record[$kDocCategoryListIndex];
			my($count) = $count = &SListLength($parents, $kValueSeperator);
			if ($count != 1) {
				$parents = &SListRemove($parents, $category, $kValueSeperator);
				$record[$kDocCategoryListIndex] = $parents;
				$result = &DBReplaceRecord(\@db, $key, \@record);
			}
			else {
				$result = &DBRemoveRecord(\@db, $key);
			}
		}
	}
	&DBClose(\@db);
	return $result;
}

# --------------------------------------------------------------------------------
# OPAddGlossary($term, $defn)
# Adds a glossary record to its database
# Args:
#	$term 		(string): the glossary term.
#	$defn		(string): a definition for the term.
#							
#	Return Value(boolean): 0 if add successful, $errDuplicateGlossary if already 
#							glossary by the same name exists, 
#							$errOPFailed otherwise
sub OPAddGlossary
{
	my($term, $defn) = @_;
	my($result);
	my(@db) = ();
	$result = &DBOpen($kGlossaryDB, $kDocDBPath, $kDBWriteAccess, \@db);

	# acquire write lock
	open(LOCKFILE, ">$kGlossaryLockFile");
	&GetWriteLockOnFile("LOCKFILE") || &LogMsg("GetWriteLockOnFile failed on file handle $kGlossaryLockFile: $!\n");

	if ($result == $errNone) {
		my(@record) = ($term, $defn);
		$result = &DBAppendRecordIfNotExists(\@db, \@record);
	}
	# release write lock
	close(LOCKFILE);
	&UnlockFile("LOCKFILE") || &LogMsg("UnlockFile failed on file handle $kGlossaryLockFile: $!\n");
	
	&DBClose(\@db);
	return $result;
}


# --------------------------------------------------------------------------------
# OPRemoveGlossary($term)
# Remove the given glossary from its database
# Args:
#	$term 		(string): the glossary term.
#							
#	Return Value(boolean): 0 if remove successful, $errEntryNotFound if  
#							glossary by the same name does not exists, 
#							$errOPFailed otherwise
sub OPRemoveGlossary
{
	my($term) = @_;
	my($result);
	my(@db) = ();
	$result = &DBOpen($kGlossaryDB, $kDocDBPath, $kDBWriteAccess, \@db);

	# acquire write lock
	open(LOCKFILE, ">$kGlossaryLockFile");
	&GetWriteLockOnFile("LOCKFILE") || &LogMsg("GetWriteLockOnFile failed on file handle $kGlossaryLockFile: $!\n");

	if ($result == $errNone) {
		$result = &DBRemoveRecord(\@db, $term);
	}

	# release write lock
	close(LOCKFILE);
	&UnlockFile("LOCKFILE") || &LogMsg("UnlockFile failed on file handle $kGlossaryLockFile: $!\n");

	&DBClose(\@db);
	return $result;
}

# --------------------------------------------------------------------------------
# OPGetGlossary($key, @glossary)
# Retreive the glossary record 
# Args:
#	$key 		(string): the key for the glossary record.
#	@glossary	(list): glossary record returned as a list.					

#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPGetGlossary
{
	local($key, *glossary) = @_;
	my($result);
	my(@db) = ();
	$result = &DBOpen($kGlossaryDB, $kDocDBPath, $kDBReadAccess, \@db);
	if ($result == $errNone) {
		$result = &DBRetreiveRecord(\@db, $key, *glossary);
		&DBClose(\@db);
	}
	return $result;
}

# --------------------------------------------------------------------------------
# OPSetGlossary($key, @glossary)
# Sets the glossary record for specified $key.
# Args:
#	$key 	(string): the key for the glossary record.
#	@glossary	(list): glossary record value as a list. Here the fields not 
#						needing to be set should have 'undefined' as value.

#	Return Value(boolean): 0 if set successful, 1 if not.
sub OPSetGlossary
{
	local($key, *glossary) = @_;
	my($i, $count);
	$count = $#glossary + 1;
	my($result);
	my(@record) = ();
	my(@db) = ();
	$result = &DBOpen($kGlossaryDB, $kDocDBPath, $kDBWriteAccess, \@db);

	# acquire write lock
	open(LOCKFILE, ">$kGlossaryLockFile");
	&GetWriteLockOnFile("LOCKFILE") || &LogMsg("GetWriteLockOnFile failed on file handle $kGlossaryLockFile: $!\n");	
	
	$result = &DBRetreiveRecord(\@db, $key, \@record);
	if ($result == $errNone) {
		for ($i = 0; $i < $count; $i++) {
			if (defined($glossary[$i])) {
				$record[$i] = $glossary[$i];
			}
		}
		if ($key eq $glossary[$kGlossaryTermIndex]) {
			$result = &DBReplaceRecord(\@db, $key, \@record);
		}
		else {
			$result = &DBAppendRecordIfNotExists(\@db, \@record);
			$result = $result || &DBRemoveRecord(\@db, $key);
		}
		&DBClose(\@db);
	}
	# release write lock
	close(LOCKFILE);
	&UnlockFile("LOCKFILE") || &LogMsg("UnlockFile failed on file handle $kGlossaryLockFile: $!\n");

	return $result;
}


# --------------------------------------------------------------------------------
# OPGetAllGlossaries(@glossaryList)
# Retreive all the glossary records.
# Args:
#	@glossaryList 	(list): glossary records returned as a list.

#	Return Value(boolean): 0 if Retreive successful, 1 if not.
sub OPGetAllGlossaries
{
    local(*glossaryList) = @_;
    my($result);
    my(@db) = ();
    $result = &DBOpen($kGlossaryDB, $kDocDBPath, $kDBReadAccess, \@db);
    if ($result == $errNone) {
	my(@glossary) = ();
	$result = &DBReadFirstRecord(\@db, \@glossary);
	my($i) = 0;
	while (!$result) {
	    $glossaryList[$i++] = &SListMake(\@glossary, $kFieldSeperator);
	    $result = &DBReadNextRecord(\@db, \@glossary);
	}
    }
    &DBClose(\@db);
    return $result;	
}


# --------------------------------------------------------------------------------
# 
sub OPGetIDFromURL
{
    my($url) = @_;
    my($id) = 0;
    my(@db) = ();
    my(@record);
	my($result) = &DBOpen($kURLIDMapDB, $kDocDBPath, $kDBReadAccess, \@db);
    if ($result == $errNone) {
		$result = &DBRetreiveRecord(\@db, $url, \@record);
		if ($result == $errNone) {
			$id = $record[$kURLMapIDIndex];
		}
		&DBClose(\@db);
    }
    return $id;
}

# --------------------------------------------------------------------------------
# 
sub OPAddURLToMap
{
    my($url) = @_;
    my($result);
    my(@db) = ();
    my($urlid) = 0;
    $result = &DBOpen($kURLIDMapDB, $kDocDBPath, $kDBWriteAccess, \@db);
    if ($result == $errNone) {
	$urlid = &GetAndIncrementMasterID($kMasterURLIDFile);
	if ($urlid != 0) {
	    my(@record) = ($url, $urlid); 
	    $result = &DBAppendRecord(\@db, \@record);
	}
	&DBClose(\@db);
    }
    return $urlid;
}


# --------------------------------------------------------------------------------
# 
sub OPAddDocIdToURLIdMap
{
    my($docID, $url, $isID) = @_;
    my($urlid);
    if (!$isID) {
		$urlid = &OPGetIDFromURL($url);
		if (!$urlid) {
			$urlid = &OPAddURLToMap($url);
		}
    }
    else {
		$urlid = $url;
    }
    my($result) = 1;
    if ($urlid) {
		my(@urecord);
		my(@db) = ();
		$result = &DBOpen($kIDDocsMapDB, $kDocDBPath, $kDBWriteAccess, \@db);
		if ($result == $errNone) {
			$result = &DBRetreiveRecord(\@db, $urlid, \@urecord);
			if ($result == $errNone) {
				my($doclist) = $urecord[$kIDMapDocIDsIndex];
				$urecord[$kIDMapDocIDsIndex] = &SListAppend($doclist, $docID, $kValueSeperator);
				$result = &DBReplaceRecord(\@db, $urlid, \@urecord);
			}
			else {
				@urecord = ($urlid, $docID);
				$result = &DBAppendRecord(\@db, \@urecord);
			}
			&DBClose(\@db);
		}
    }
    return $result;
}

# --------------------------------------------------------------------------------
# 
sub OPRemoveDocIdFromURLIdMap
{
    my($docID, $url, $isID) = @_;
    my($urlid) = (!$isID) ? &OPGetIDFromURL($url) : $url;

    my($result) = 1;
    if ($urlid) {
		my(@record);
		my(@db) = ();
		$result = &DBOpen($kIDDocsMapDB, $kDocDBPath, $kDBWriteAccess, \@db);
		if ($result == $errNone) {
			$result = &DBRetreiveRecord(\@db, $urlid, \@record);
			if ($result == $errNone) {
				my($doclist) = $record[$kIDMapDocIDsIndex];
				$record[$kIDMapDocIDsIndex] = &SListRemove($doclist, $docID, $kValueSeperator);
				$result = &DBReplaceRecord(\@db, $urlid, \@record);
			}
			&DBClose(\@db);
		}
    }
    return $result;
}

# --------------------------------------------------------------------------------
# 
sub OPGetDocsForURL
{
    local($url, $isID, *docs) = @_;
    my($result);
    my(@db) = ();
    my(@record);
    @docs = ();

    my($urlid) = (!$isID) ? &OPGetIDFromURL($url) : $url;

    if ($urlid) {
		&DBOpen($kIDDocsMapDB, $kDocDBPath, $kDBReadAccess, \@db);
		if ($result == $errNone) {
			$result = &DBRetreiveRecord(\@db, $urlid, \@record);
			@docs = split($kValueSeperator, $record[$kIDMapDocIDsIndex]);
			&DBClose(\@db);
		}
    }
    return $#docs + 1;
}

1;
